class LocustNotFoundException(Exception):
    pass


class LoadDensityTestException(Exception):
    pass


# json
class LoadDensityTestJsonException(LoadDensityTestException):
    pass


class LoadDensityGenerateJsonReportException(LoadDensityTestException):
    pass


class LoadDensityTestExecuteException(LoadDensityTestException):
    pass


class LoadDensityAssertException(LoadDensityTestException):
    pass


class LoadDensityHTMLException(LoadDensityTestException):
    pass


class LoadDensityAddCommandException(LoadDensityTestException):
    pass


# XML

class XMLException(LoadDensityTestException):
    pass


class XMLTypeException(XMLException):
    pass


# Execute callback
class CallbackExecutorException(LoadDensityTestException):
    pass



traditional_chinese_word_dict = {
    # Main
    "application_name": "LoadDensity",
    # Widget
    "url": "URL: ",
    "test_time": "測試時間: ",
    "user_count": "測試使用者總數: ",
    "spawn_rate": "產生速率: ",
    "test_method": "測試 HTTP 方法: ",
    "start_button": "開始: ",
    "log": "日誌: ",
    # Test method
    "get": "GET",
    "post": "POST",
    "put": "PUT",
    "patch": "PATCH",
    "delete": "DELETE",
    "head": "HEAD",
    "options": "OPTIONS",
}



# locust error
not_found_locust_error: str = "Locust not found"

# json error
cant_reformat_json_error: str = "can't reformat JSON: is the type correct?"
wrong_json_data_error: str = "can't parse JSON"
cant_find_json_error: str = "can't find JSON file"
cant_save_json_error: str = "can't save JSON file"
cant_generate_json_report: str = "can't generate JSON report"

# executor error
executor_data_error: str = "executor received invalid data"
executor_list_error: str = "executor received invalid data: list is empty or wrong type"

# HTML
html_generate_no_data_tag: str = "no records to generate HTML report"

# add command
add_command_exception_tag: str = "command value type must be a method or function"

# argparse
argparse_get_wrong_data: str = "argparse received invalid data"

# XML
cant_read_xml_error: str = "can't read XML"
xml_type_error: str = "XML type error"

# Callback executor
get_bad_trigger_method: str = "invalid trigger method: only kwargs and args accepted"
get_bad_trigger_function: str = "invalid trigger function: only functions in event_dict accepted"


import json
import sys
from socket import AF_INET, SOCK_STREAM

import gevent
from gevent import monkey
from gevent import socket

from je_load_density.utils.executor.action_executor import execute_action


class TCPServer(object):

    def __init__(self):
        self.close_flag: bool = False
        self.server: socket.socket = socket.socket(AF_INET, SOCK_STREAM)

    def socket_server(self, host: str, port: int):
        self.server.bind((host, port))
        self.server.listen()
        while not self.close_flag:
            connection = self.server.accept()[0]
            gevent.spawn(self.handle, connection)
        sys.exit(0)

    def handle(self, connection):
        connection_data = connection.recv(8192)
        command_string = str(connection_data.strip(), encoding="utf-8")
        print("command is: " + command_string, flush=True)
        if command_string == "quit_server":
            connection.close()
            self.close_flag = True
            self.server.close()
            print("Now quit server", flush=True)
        else:
            try:
                execute_str = json.loads(command_string)
                if execute_str is not None:
                    for execute_return in execute_action(execute_str).values():
                        connection.send(str(execute_return).encode("utf-8"))
                        connection.send("\n".encode("utf-8"))
                connection.send("Return_Data_Over_JE".encode("utf-8"))
                connection.send("\n".encode("utf-8"))
            except Exception as error:
                try:
                    connection.send(str(error).encode("utf-8"))
                    connection.send("\n".encode("utf-8"))
                    connection.send("Return_Data_Over_JE".encode("utf-8"))
                    connection.send("\n".encode("utf-8"))
                except Exception as error:
                    print(repr(error))
                    sys.exit(1)
            finally:
                connection.close()


def start_load_density_socket_server(host: str = "localhost", port: int = 9940):
    monkey.patch_all()
    server = TCPServer()
    server.socket_server(host, port)
    return server



english_word_dict = {
    # Main
    "application_name": "LoadDensity",
    # Widget
    "url": "URL",
    "test_time": "Test Time: ",
    "user_count": "User Count: ",
    "spawn_rate": "Spawn Rate: ",
    "test_method": "Test Method: ",
    "start_button": "Start: ",
    "log": "Log: ",
    # Test method
    "get": "GET",
    "post": "POST",
    "put": "PUT",
    "patch": "PATCH",
    "delete": "DELETE",
    "head": "HEAD",
    "options": "OPTIONS",
}



from je_load_density.utils.logging.loggin_instance import load_density_logger
from je_load_density.wrapper.create_locust_env.create_locust_env import prepare_env
from je_load_density.wrapper.user_template.fast_http_user_template import FastHttpUserWrapper, set_wrapper_fasthttp_user
from je_load_density.wrapper.user_template.http_user_template import HttpUserWrapper, set_wrapper_http_user


def start_test(
        user_detail_dict: dict,
        user_count: int = 50, spawn_rate: int = 10, test_time: int = 60,
        web_ui_dict: dict = None,
        **kwargs
):
    """
    :param user_detail_dict: dict use to create user
    :param user_count: how many user we want to spawn
    :param spawn_rate: one time will spawn how many user
    :param test_time: total test run time
    :param web_ui_dict: web ui dict include host and port like {"host": "127.0.0.1", "port": 8089}
    :param kwargs: to catch unknown param
    :return: None
    """
    load_density_logger.info(
        f"start_test, user_detail_dict: {user_detail_dict}, user_count: {user_count}, "
        f"spawn_rate: {spawn_rate}, test_time: {test_time}, web_ui_dict: {web_ui_dict}, "
        f"params: {kwargs}"
    )
    user_dict = {
        "fast_http_user": {"actually_user": FastHttpUserWrapper, "init": set_wrapper_fasthttp_user},
        "http_user": {"actually_user": HttpUserWrapper, "init": set_wrapper_http_user}
    }
    user = user_dict.get(user_detail_dict.get("user", "fast_http_user"))
    actually_user = user.get("actually_user", "actually_user")
    init_function = user.get("init", "init")
    init_function(user_detail_dict, **kwargs)

    prepare_env(
        user_class=actually_user, user_count=user_count, spawn_rate=spawn_rate, test_time=test_time,
        web_ui_dict=web_ui_dict, **kwargs
    )
    return str(user_detail_dict) + " " + "user_count: " + str(user_count) + " spawn_rate: " + str(
        spawn_rate) + " test_time: " + str(test_time)



# hook
from je_load_density.wrapper.event.request_hook import request_hook
# env
from je_load_density.utils.executor.action_executor import add_command_to_executor
# executor
from je_load_density.utils.executor.action_executor import execute_action
from je_load_density.utils.executor.action_executor import execute_files
from je_load_density.utils.executor.action_executor import executor
# file
from je_load_density.utils.file_process.get_dir_file_list import get_dir_files_as_list
# html
from je_load_density.utils.generate_report.generate_html_report import generate_html
from je_load_density.utils.generate_report.generate_html_report import generate_html_report
# json
from je_load_density.utils.generate_report.generate_json_report import generate_json
from je_load_density.utils.generate_report.generate_json_report import generate_json_report
from je_load_density.utils.json.json_file.json_file import read_action_json
# xml
from je_load_density.utils.generate_report.generate_xml_report import generate_xml
from je_load_density.utils.generate_report.generate_xml_report import generate_xml_report
# server
from je_load_density.utils.socket_server.load_density_socket_server import start_load_density_socket_server
# test record
from je_load_density.utils.test_record.test_record_class import test_record_instance
# start
from je_load_density.wrapper.create_locust_env.create_locust_env import prepare_env
from je_load_density.wrapper.create_locust_env.create_locust_env import create_env

# Proxy
from je_load_density.wrapper.proxy.proxy_user import locust_wrapper_proxy

from je_load_density.wrapper.start_wrapper.start_test import start_test

# Locust
from locust import SequentialTaskSet
from locust import task
from locust import TaskSet

# Callback
from je_load_density.utils.callback.callback_function_executor import callback_executor

from je_load_density.utils.project.create_project_structure import create_project_dir

__all__ = [
    "create_env", "start_test",
    "locust_wrapper_proxy",
    "prepare_env", "prepare_env",
    "test_record_instance",
    "execute_action", "execute_files", "executor", "add_command_to_executor",
    "get_dir_files_as_list",
    "generate_html", "generate_html_report",
    "generate_json", "generate_json_report", "read_action_json",
    "generate_xml", "generate_xml_report",
    "start_load_density_socket_server",
    "SequentialTaskSet", "task", "TaskSet",
    "callback_executor", "create_project_dir"
]



import logging
import queue

locust_log_queue = queue.Queue()

class InterceptAllFilter(logging.Filter):

    def filter(self, record):
        locust_log_queue.put(record.getMessage())
        return False



import json
import sys
from threading import Lock

from je_load_density.utils.exception.exception_tags import cant_generate_json_report
from je_load_density.utils.exception.exceptions import LoadDensityGenerateJsonReportException
from je_load_density.utils.test_record.test_record_class import test_record_instance


def generate_json():
    if len(test_record_instance.test_record_list) == 0 and len(test_record_instance.error_record_list) == 0:
        raise LoadDensityGenerateJsonReportException(cant_generate_json_report)
    else:
        success_dict = dict()
        failure_dict = dict()
        failure_count: int = 1
        failure_test_str: str = "Failure_Test"
        success_count: int = 1
        success_test_str: str = "Success_Test"
        for record_data in test_record_instance.test_record_list:
            success_dict.update(
                {
                    success_test_str + str(success_count): {
                        "Method": str(record_data.get("Method")),
                        "test_url": str(record_data.get("test_url")),
                        "name": str(record_data.get("name")),
                        "status_code": str(record_data.get("status_code")),
                        "text": str(record_data.get("text")),
                        "content": str(record_data.get("content")),
                        "headers": str(record_data.get("headers"))
                    }
                }
            )
            success_count = success_count + 1
        for record_data in test_record_instance.error_record_list:
            failure_dict.update(
                {
                    failure_test_str + str(failure_count): {
                        "Method": str(record_data.get("Method")),
                        "test_url": str(record_data.get("test_url")),
                        "name": str(record_data.get("name")),
                        "status_code": str(record_data.get("status_code")),
                        "error": str(record_data.get("error"))
                    }
                }
            )
            failure_count = failure_count + 1
    return success_dict, failure_dict


def generate_json_report(json_file_name: str = "default_name"):
    """
    :param json_file_name: save json file's name
    """
    lock = Lock()
    success_dict, failure_dict = generate_json()
    try:
        lock.acquire()
        with open(json_file_name + "_success.json", "w+") as file_to_write:
            json.dump(dict(success_dict), file_to_write, indent=4)
    except Exception as error:
        print(repr(error), file=sys.stderr)
    finally:
        lock.release()
    try:
        lock.acquire()
        with open(json_file_name + "_failure.json", "w+") as file_to_write:
            json.dump(dict(failure_dict), file_to_write, indent=4)
    except Exception as error:
        print(repr(error), file=sys.stderr)
    finally:
        lock.release()



import logging

from PySide6.QtCore import QTimer
from PySide6.QtGui import QIntValidator
from PySide6.QtWidgets import QWidget, QFormLayout, QLineEdit, QComboBox, QPushButton, QTextEdit, QVBoxLayout, QLabel

from je_load_density.gui.load_density_gui_thread import LoadDensityGUIThread
from je_load_density.gui.language_wrapper.multi_language_wrapper import language_wrapper
from je_load_density.gui.log_to_ui_filter import InterceptAllFilter, locust_log_queue


class LoadDensityWidget(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        # from
        form_layout = QFormLayout()
        self.url_input = QLineEdit()
        self.test_time_input = QLineEdit()
        self.test_time_input.setValidator(QIntValidator())
        self.user_count_input = QLineEdit()
        self.user_count_input.setValidator(QIntValidator())
        self.spawn_rate_input = QLineEdit()
        self.spawn_rate_input.setValidator(QIntValidator())
        self.method_combobox = QComboBox()
        self.method_combobox.addItems([
            language_wrapper.language_word_dict.get("get"),
            language_wrapper.language_word_dict.get("post"),
            language_wrapper.language_word_dict.get("put"),
            language_wrapper.language_word_dict.get("patch"),
            language_wrapper.language_word_dict.get("delete"),
            language_wrapper.language_word_dict.get("head"),
            language_wrapper.language_word_dict.get("options"),
        ])
        form_layout.addRow(language_wrapper.language_word_dict.get("url"), self.url_input)
        form_layout.addRow(language_wrapper.language_word_dict.get("test_time"), self.test_time_input)
        form_layout.addRow(language_wrapper.language_word_dict.get("user_count"), self.user_count_input)
        form_layout.addRow(language_wrapper.language_word_dict.get("spawn_rate"), self.spawn_rate_input)
        form_layout.addRow(language_wrapper.language_word_dict.get("test_method"), self.method_combobox)

        self.start_button = QPushButton(language_wrapper.language_word_dict.get("start_button"))
        self.start_button.clicked.connect(self.run_load_density)

        # Log panel
        self.log_panel = QTextEdit()
        self.log_panel.setReadOnly(True)

        # Add widget to vertical layout
        main_layout = QVBoxLayout()
        main_layout.addLayout(form_layout)
        main_layout.addWidget(self.start_button)
        main_layout.addWidget(QLabel(language_wrapper.language_word_dict.get("log")))
        main_layout.addWidget(self.log_panel)

        # Param
        self.run_load_density_thread = None
        self.pull_log_timer = QTimer()
        self.pull_log_timer.setInterval(20)
        self.pull_log_timer.timeout.connect(self.add_text_to_log)

        self.setLayout(main_layout)

    def run_load_density(self):
        self.run_load_density_thread = LoadDensityGUIThread()
        self.run_load_density_thread.url = self.url_input.text()
        self.run_load_density_thread.test_time = int(self.test_time_input.text())
        self.run_load_density_thread.user_count = int(self.user_count_input.text())
        self.run_load_density_thread.spawn_rate = int(self.spawn_rate_input.text())
        self.run_load_density_thread.method = self.method_combobox.currentText().lower()
        log_handler_list = [handler for handler in logging.getLogger("root").handlers if handler.name == "log_reader"]
        if log_handler_list:
            log_handler = log_handler_list[0]
            log_handler.addFilter(InterceptAllFilter())
        self.run_load_density_thread.start()
        self.pull_log_timer.stop()
        self.pull_log_timer.start()
        self.log_panel.clear()

    def add_text_to_log(self):
        if not locust_log_queue.empty():
            self.log_panel.append(locust_log_queue.get_nowait())



from PySide6.QtCore import QThread
from je_load_density.wrapper.start_wrapper.start_test import start_test


class LoadDensityGUIThread(QThread):

    def __init__(self):
        super().__init__()
        self.url = None
        self.test_time = None
        self.user_count = None
        self.spawn_rate = None
        self.method = None

    def run(self):
        start_test(
            {
                "user": "fast_http_user",
            },
            self.user_count, self.spawn_rate, self.test_time,
            **{
                "tasks": {
                    self.method: {"request_url": self.url},
                }
            }
        )